package fstree

import (
	"context"
	"errors"
	"testing"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/shard/mode"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/util/logicerr"
	objectSDK "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object"
	oidtest "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id/test"
	"github.com/stretchr/testify/require"
	"golang.org/x/sync/errgroup"
)

func TestAddressToString(t *testing.T) {
	addr := oidtest.Address()
	s := stringifyAddress(addr)
	actual, err := addressFromString(s)
	require.NoError(t, err)
	require.Equal(t, addr, actual)
}

func Benchmark_addressFromString(b *testing.B) {
	addr := oidtest.Address()
	s := stringifyAddress(addr)

	b.ReportAllocs()
	b.ResetTimer()
	for range b.N {
		_, err := addressFromString(s)
		if err != nil {
			b.Fatalf("benchmark error: %v", err)
		}
	}
}

func TestObjectCounter(t *testing.T) {
	t.Parallel()
	counter := NewSimpleCounter()
	fst := New(
		WithPath(t.TempDir()),
		WithDepth(2),
		WithDirNameLen(2),
		WithFileCounter(counter))
	require.NoError(t, fst.Open(mode.ComponentReadWrite))
	require.NoError(t, fst.Init())

	count, size := counter.CountSize()
	require.Equal(t, uint64(0), count)
	require.Equal(t, uint64(0), size)

	defer func() {
		require.NoError(t, fst.Close())
	}()

	addr := oidtest.Address()
	obj := objectSDK.New()
	obj.SetID(addr.Object())
	obj.SetContainerID(addr.Container())
	obj.SetPayload([]byte{1, 2, 3, 4, 5, 6, 7, 8, 9, 0})

	var putPrm common.PutPrm
	putPrm.Address = addr
	putPrm.RawData, _ = obj.Marshal()

	var delPrm common.DeletePrm
	delPrm.Address = addr

	t.Run("without size hint", func(t *testing.T) {
		eg, egCtx := errgroup.WithContext(context.Background())

		eg.Go(func() error {
			for range 1_000 {
				_, err := fst.Put(egCtx, putPrm)
				if err != nil {
					return err
				}
			}
			return nil
		})

		eg.Go(func() error {
			var le logicerr.Logical
			for range 1_000 {
				_, err := fst.Delete(egCtx, delPrm)
				if err != nil && !errors.As(err, &le) {
					return err
				}
			}
			return nil
		})

		require.NoError(t, eg.Wait())

		count, size = counter.CountSize()
		realCount, realSize, err := fst.countFiles()
		require.NoError(t, err)
		require.Equal(t, realCount, count, "real %d, actual %d", realCount, count)
		require.Equal(t, realSize, size, "real %d, actual %d", realSize, size)
	})

	t.Run("with size hint", func(t *testing.T) {
		delPrm.Size = uint64(len(putPrm.RawData))
		eg, egCtx := errgroup.WithContext(context.Background())

		eg.Go(func() error {
			for range 1_000 {
				_, err := fst.Put(egCtx, putPrm)
				if err != nil {
					return err
				}
			}
			return nil
		})

		eg.Go(func() error {
			var le logicerr.Logical
			for range 1_000 {
				_, err := fst.Delete(egCtx, delPrm)
				if err != nil && !errors.As(err, &le) {
					return err
				}
			}
			return nil
		})

		require.NoError(t, eg.Wait())

		count, size = counter.CountSize()
		realCount, realSize, err := fst.countFiles()
		require.NoError(t, err)
		require.Equal(t, realCount, count, "real %d, actual %d", realCount, count)
		require.Equal(t, realSize, size, "real %d, actual %d", realSize, size)
	})
}
