package writecache

import (
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/logger"
	"go.uber.org/zap"
)

// Option represents write-cache configuration option.
type Option func(*options)

type options struct {
	log *logger.Logger
	// path is a path to a directory for write-cache.
	path string
	// blobstor is the main persistent storage.
	blobstor MainStorage
	// metabase is the metabase instance.
	metabase Metabase
	// maxObjectSize is the maximum size of the object stored in the write-cache.
	maxObjectSize uint64
	// workersCount is the number of workers flushing objects in parallel.
	workersCount int
	// maxCacheSize is the maximum total size of all objects saved in cache (DB + FS).
	// 1 GiB by default.
	maxCacheSize uint64
	// maxCacheCount is the maximum total count of all object saved in cache.
	// 0 (no limit) by default.
	maxCacheCount uint64
	// noSync is true iff FSTree allows unsynchronized writes.
	noSync bool
	// reportError is the function called when encountering disk errors in background workers.
	reportError func(string, error)
	// metrics is metrics implementation
	metrics Metrics
	// disableBackgroundFlush is for testing purposes only.
	disableBackgroundFlush bool
	// flushSizeLimit is total size of flushing objects.
	flushSizeLimit uint64
}

// WithLogger sets logger.
func WithLogger(log *logger.Logger) Option {
	return func(o *options) {
		o.log = &logger.Logger{Logger: log.With(zap.String("component", "WriteCache"))}
	}
}

// WithPath sets path to writecache db.
func WithPath(path string) Option {
	return func(o *options) {
		o.path = path
	}
}

// WithBlobstor sets main object storage.
func WithBlobstor(bs MainStorage) Option {
	return func(o *options) {
		o.blobstor = bs
	}
}

// WithMetabase sets metabase.
func WithMetabase(db Metabase) Option {
	return func(o *options) {
		o.metabase = db
	}
}

// WithMaxObjectSize sets maximum object size to be stored in write-cache.
func WithMaxObjectSize(sz uint64) Option {
	return func(o *options) {
		if sz > 0 {
			o.maxObjectSize = sz
		}
	}
}

func WithFlushWorkersCount(c int) Option {
	return func(o *options) {
		if c > 0 {
			o.workersCount = c
		}
	}
}

// WithMaxCacheSize sets maximum write-cache size in bytes.
func WithMaxCacheSize(sz uint64) Option {
	return func(o *options) {
		o.maxCacheSize = sz
	}
}

// WithMaxCacheCount sets maximum write-cache objects count.
func WithMaxCacheCount(v uint64) Option {
	return func(o *options) {
		o.maxCacheCount = v
	}
}

// WithNoSync sets an option to allow returning to caller on PUT before write is persisted.
// Note, that we use this flag for FSTree only and DO NOT use it for a bolt DB because
// we cannot yet properly handle the corrupted database during the startup. This SHOULD NOT
// be relied upon and may be changed in future.
func WithNoSync(noSync bool) Option {
	return func(o *options) {
		o.noSync = noSync
	}
}

// WithReportErrorFunc sets error reporting function.
func WithReportErrorFunc(f func(string, error)) Option {
	return func(o *options) {
		o.reportError = f
	}
}

// WithMetrics sets metrics implementation.
func WithMetrics(metrics Metrics) Option {
	return func(o *options) {
		o.metrics = metrics
	}
}

// WithDisableBackgroundFlush disables background flush, for testing purposes only.
func WithDisableBackgroundFlush() Option {
	return func(o *options) {
		o.disableBackgroundFlush = true
	}
}

// WithFlushSizeLimit sets flush size limit.
func WithFlushSizeLimit(v uint64) Option {
	return func(o *options) {
		o.flushSizeLimit = v
	}
}
